/*
 * @DEC_COPYRIGHT@
 */
/*
 * HISTORY
 * $Log: strcpy.S,v $
# Revision 1.1  1994/04/01  19:52:08  paradis
# Initial revision
#
 * Revision 1.1.2.4  1993/01/19  14:06:12  Ken_Lesniak
 * 	add .prologue directive
 * 	[93/01/08  17:19:15  Ken_Lesniak]
 *
 * Revision 1.1.2.3  92/08/21  14:11:46  Ken_Lesniak
 * 	fix so we don't access past the end of strings in the unaligned case
 * 	[92/08/19  16:28:25  Ken_Lesniak]
 * 
 * Revision 1.1.2.2  92/06/30  14:45:48  Ken_Lesniak
 * 	replace strcpy.c
 * 	[92/06/30  14:31:18  Ken_Lesniak]
 * 
 * $EndLog$
 */
/*
 * @(#)$RCSfile: strcpy.S,v $ $Revision: 1.1 $ (DEC) $Date: 1994/04/01 19:52:08 $
 */

#include <linux/alpha/regdef.h>

/*
 * extern char *strcpy(char *s1, const char *s2);
 *
 *	a0	dst string pointer (s1)
 *	a1	src string pointer (s2)
 *
 * Copy the string pointed to by the "s2" parameter, including the
 * terminating null character, to the location pointed to by the
 * "s1" parameter. When operating on overlapping strings, the
 * behavior is unreliable.
 *
 * The return value contains a pointer to the destination string (s1).
 */
	.globl	strcpy
	.ent	strcpy

	ldq_u	t0, 0(a1)		# get first src quad
	xor	a1, a0, t4		# see where src and dst differ
	bis	a0, a0, v0		# set return value
	and	t4, 8-1, t4		# only interested in low 3 bits
	and	a0, 8-1, t7		# get offset to first dst byte
	bne	t4, unaligned		# skip if unaligned
	cmpbge	zero, t0, t1		# any null bytes?
	bne	t7, initial		# copy initial bytes
	nop				# improve scheduling
	bne	t1, final		# br if found null byte

	# The src and dst pointers are quad aligned, so copy a
	# quadword at a time until we find one with a null byte
	#
	#	a0	dst pointer (not necessarily aligned)
	#	a1	src pointer (not necessarily aligned)
	#	t0	8 src bytes pointed to by a1

	.align	3
quads:	stq_u	t0, 0(a0)		# store next quad
	addq	a0, 8, a0		# advance dst pointer
	ldq_u	t0, 8(a1)		# get next src quad
	cmpbge	zero, t0, t1		# any null bytes?
	addq	a1, 8, a1		# advance src pointer
	beq	t1, quads		# continue looping if not

	# We loaded a quad containing a null byte -- copy
	# only up to the null byte
	#
	#	a0	dst pointer
	#	t0	src quadword containing null byte
	#	t1	bit mask of null bytes within t0

	.align	3
final:	ldq_u	t2, 0(a0)		# get dst quad
	subq	zero, t1, t3		# flip bits left of first bit set
	xor	t1, t3, t1		# bit mask of bytes after null
	zap	t0, t1, t0		# zap src bytes to ignore
	zapnot	t2, t1, t2		# zap dst bytes to ignore
	nop				# improve scheduling
	or	t0, t2, t0		# merge src and dst bytes
	stq_u	t0, 0(a0)		# store the merged bytes
	ret	zero,(ra)

	# The src and dst pointers have the same alignment, but the
	# initial pointers are not quad aligned. Copy to the end of
	# the quad or to the null byte if there is one
	#
	#	a0	dst pointer
	#	a1	src pointer
	#	t0	src quadword containing null byte
	#	t1	bit mask of null bytes within t0
	#	t7	dst pointer alignment

	.align	3
initial:
	ldq_u	t2, 0(a0)		# get dst quad
	ldiq	t6, 255			# init bit mask to all bytes
	sll	t6, t7, t6		# get bit mask of relevant bytes
	and	t6, t1, t1		# ignore nulls before start of string
	subq	zero, t1, t3		# flip bits left of first bit set
	xor	t1, t3, t3		# bit mask of bytes after null
	ornot	t3, t6, t6		# bit mask of non-src bytes
	zap	t0, t6, t0		# zap src bytes to ignore
	zapnot	t2, t6, t2		# zap dst bytes to ignore
	nop				# improve scheduling
	or	t0, t2, t0		# merge src and dst bytes
	beq	t3, quads		# copy quads if we did not find a null
	stq_u	t0, 0(a0)		# store the merged bytes
	ret	zero,(ra)	

	# The src and dst pointers do not have the same alignment.
	#
	#	a0	dst pointer
	#	a1	src pointer
	#	t0	initial source bytes
	#	t7	dst pointer alignment

	.align	3
unaligned:
	subq	zero, a0, t5		# negate dst pointer
	and	t5, 8-1, t5		# bytes till dst is aligned
	ldiq	t6, 255			# bit mask of 8 bytes all 1
	bne	t5, uinitial		# br if dst is not already aligned

	# The next dst byte is on a quadword boundary, so try to
	# copy a quadword at a time. Continue copying until we find
	# a source byte with a null.
	#
	#	a0	dst pointer
	#	a1	src pointer
	#	t0	initial source bytes
	#	t6	all-byte bit mask

	.align	3
uquads:	and	a1, 8-1, t7		# get src alignment
	srl	t6, t7, t6		# get mask of initial bytes

10:	extql	t0, a1, t8		# get low bytes of src
	cmpbge	zero, t8, t1		# any null bytes?
	and	t1, t6, t2		# only look at initial bytes
	bne	t2, ufinal		# br if we found a null

	ldq_u	t0, 8(a1)		# get next 8 src bytes
	addq	a1, 8, a1		# advance src pointer

	extqh	t0, a1, t9		# get high bytes of src
	or	t8, t9, t8		# merge high and low bytes
	cmpbge	zero, t8, t1		# any null bytes?
	bne	t1, ufinal		# br if we found a null

	stq	t8, 0(a0)		# store the next 8 bytes
	addq	a0, 8, a0		# advance the dst pointer
	br	zero, 10b		# continue till we find a null

	# We loaded a source quadword that contains a null -- copy
	# only up to the null byte
	#
	#	a0	dst pointer
	#	t8	src quadword containing null byte
	#	t1	bit mask of null bytes within t8

	.align	3
ufinal:
	bis	t8, t8, t0		# copy src bytes containing null
	br	zero, final		# continue with aligned case

	# The src and dst pointers do not have the same alignment,
	# and the dst pointer is not quadword aligned. Copy the
	# initial bytes so that the dst pointer is quadword aligned,
	# and then copy a quadword at a time.
	#
	#	a0	dst pointer
	#	a1	src pointer
	#	t0	initial source bytes
	#	t5	bytes till dst is QW aligned
	#	t6	all-byte bit mask
	#	t7	dst pointer alignment

	.align	3
uinitial:
	ldq_u	t2, 0(a0)		# get dst quadword
	ldiq	t3, -1			# get 8 bytes of 0xff
	extql	t3, a0, t4		# get mask of the dst bytes
	extql	t3, a1, t3		# get mask of the low src bytes
	addq	a1, t5, a3		# get pointer past initial src bytes

	extql	t0, a1, t9		# get the low initial src bytes
	ornot	t9, t4, t9		# set non-dst bytes to 0xff
	ornot	t9, t3, t8		# set non-src bytes to 0xff
	cmpbge	zero, t8, t1		# any null bytes?
	bne	t1, upartial		# br if a null found

	mskql	t2, a0, t4		# zap initial dst bytes

	ldq_u	t0, -1(a3)		# get high bytes of initial src bytes
	extqh	t0, a1, t8		# extract the high src bytes
	or	t8, t9, t8		# merge the high and low src bytes
	cmpbge	zero, t8, t1		# any null bytes?
	insql	t8, a0, t3		# shift src to match dst alignment
	addq	a1, t5, a1		# advance the src pointer
	bne	t1, upartial		# br if a null found

	or	t4, t3, t2		# merge src and dst bytes
	stq_u	t2, 0(a0)		# store the merged bytes
	addq	a0, t5, a0		# advance the dst pointer

	ldq_u	t0, 0(a1)		# get the next low src bytes
	br	zero, uquads		# copy by quadword now

	# The src and dst pointers do not have the same alignment,
	# and the dst pointer is not quadword aligned, and the initial
	# bytes of the src contained a null byte. Merge the src
	# bytes with the dst bytes and exit.
	#
	#	a0	dst pointer
	#	t2	initial dst bytes
	#	t6	all-byte bit mask
	#	t7	dst pointer alignment
	#	t8	initial source bytes

	.align	3
upartial:
	insql	t8, a0, t8		# shift src to match dst alignment
	sll	t1, t7, t1		# shift null-byte mask to match src
	sll	t6, t7, t6		# get mask of initial src bytes
	subq	zero, t1, t4		# flip bits after first
	xor	t1, t4, t4		# bit mask of bytes after null
	ornot	t4, t6, t6		# mask of src bytes to be copied
	zap	t8, t6, t8		# zap src bytes to ignore
	zapnot	t2, t6, t2		# zap dst bytes to ignore
	or	t2, t8, t2		# merge src and dst bytes
	stq_u	t2, 0(a0)		# store the merged bytes
	ret	zero,(ra)
	
	.end	strcpy
