/*
 * @DEC_COPYRIGHT@
 */
/*
 * HISTORY
 * $Log: __divl.S,v $
# Revision 1.1  1994/04/01  18:58:28  paradis
# Initial revision
#
 * Revision 1.2.2.9  1992/12/22  14:33:32  Ken_Lesniak
 *      fixup prolog and register usage to work with exception handling.
 *      modify divide by zero trap to return 0 if continued
 *      [92/12/21  11:39:26  Ken_Lesniak]
 *
 * Revision 1.2.2.8  92/12/08  18:03:34  Ken_Lesniak
 *      absolute value of inputs should be done in 64 bits
 *      [92/12/08  18:01:59  Ken_Lesniak]
 *
 * Revision 1.2.2.7  92/11/20  16:37:39  Ken_Lesniak
 *      rearrange so we save and restore one less register
 *      [92/11/20  16:35:44  Ken_Lesniak]
 *
 * Revision 1.2.2.6  92/11/04  15:31:37  Ken_Lesniak
 *      total re-write
 *      [92/11/04  14:44:46  Ken_Lesniak]
 *
 * Revision 1.2.2.5  92/08/26  10:48:51  Ken_Lesniak
 *      return result in $27
 *      [92/08/26  10:25:10  Ken_Lesniak]
 *
 * Revision 1.2.2.4  92/06/11  08:41:49  Mike_Rickabaugh
 *      Updated divide/mod routines to use registers , , and .
 *      [92/06/11  08:36:38  Mike_Rickabaugh]
 *
 * Revision 1.2.2.3  92/04/13  14:58:39  Mike_Rickabaugh
 *      Altered divide and friends to be prepared for both calling standards.
 *      [92/04/13  14:55:26  Mike_Rickabaugh]
 *
 * Revision 1.2.2.2  92/02/20  17:00:55  Mike_Rickabaugh
 *      Initial 3.10 version.
 *      [92/02/20  16:30:54  Mike_Rickabaugh]
 *
 * $EndLog$
 */

/*
 * Signed division
 *
 * On input:
 *      $23/t9  = return address (RA)
 *      $24/t10 = dividend (numerator) (A)
 *      $25/t11 = divisor (denominator) (B)
 *      Note that the high 32 bits of r24 and r25 are ignored.
 *
 * On Return:
 *      $27/t12 = quotient (Q)
 *      All other registers, except $at and $gp, are preserved
 */

#define ENTRY \
        .globl  __divl; \
        .ent    __divl; \
__divl: ; \
        ldgp    gp, 0(pv); \
        lda     sp, -8*8(sp); \
        stq     RA, 0*8(sp); \
        stq     t0, 1*8(sp); \
        stq     t1, 2*8(sp); \
        stq     t2, 3*8(sp); \
        stq     t3, 4*8(sp); \
        stq     a0, 5*8(sp); \
        stq     A, 6*8(sp); \
        stq     B, 7*8(sp); \
        .prologue 1; \
        addl    A, 0, A         /* make sure dividend is in canonical form */; \
        addl    B, 0, B         /* make sure divisor is in canonical form */; \
        /* check for INT_MIN / -1 */; \
        addl    B, 1, t0        /* 0 if B == -1; != 0 otherwise */; \
        bis     A, A, t1        /* copy dividend */; \
        cmovne  t0, 0, t1       /* replace w/ 0 if divisor != -1 */; \
        sublv   zero, t1, t1    /* trap if dividend = INT_MIN */; \
        subq    zero, A, t0     /* negate dividend */; \
        cmovlt  A, t0, A        /* get 64-bit absolute value of dividend */; \
        subq    zero, B, t1     /* negate divisor */; \
        cmovlt  B, t1, B        /* get 64-bit absolute value of divisor */

#define THIRTYTWOBIT 1

#define RETURN \
        addl    Q, 0, Q; \
        subl    zero, Q, AT; \
        ldq     A, 6*8(sp); \
        ldq     B, 7*8(sp); \
        xor     A, B, t0        /* compute sign of quotient */; \
        addl    t0, 0, t0; \
        cmovlt  t0, AT, Q; \
        ldq     RA, 0*8(sp); \
        ldq     t0, 1*8(sp); \
        ldq     t1, 2*8(sp); \
        ldq     t2, 3*8(sp); \
        ldq     t3, 4*8(sp); \
        ldq     a0, 5*8(sp); \
        lda     sp, 8*8(sp); \
        ret     zero, (RA), 1

#define END \
	.end	__divl

#include "divrem.S"

