/*
 * @DEC_COPYRIGHT@
 */
/*
 * HISTORY
 * $Log: __remq.s,v $
 * Revision 1.1.2.8  1992/12/22  14:34:17  Ken_Lesniak
 * 	fixup prolog and register usage to work with exception handling.
 * 	modify divide by zero trap to return 0 if continued
 * 	[92/12/21  11:40:34  Ken_Lesniak]
 *
 * Revision 1.1.2.7  92/11/20  16:38:04  Ken_Lesniak
 * 	rearrange so we save and restore one less register
 * 	[92/11/20  16:36:40  Ken_Lesniak]
 * 
 * Revision 1.1.2.6  92/11/04  15:32:31  Ken_Lesniak
 * 	total re-write
 * 	[92/11/04  14:47:52  Ken_Lesniak]
 * 
 * Revision 1.1.2.5  92/08/26  10:49:33  Ken_Lesniak
 * 	return results in $27
 * 	[92/08/26  10:37:41  Ken_Lesniak]
 * 
 * Revision 1.1.2.4  92/06/11  08:43:09  Mike_Rickabaugh
 * 	Updated divide/mod routines to use registers , , and .
 * 	[92/06/11  08:38:02  Mike_Rickabaugh]
 * 
 * Revision 1.1.2.3  92/04/13  14:59:38  Mike_Rickabaugh
 * 	Altered divide and friends to be prepared for both calling standards.
 * 	[92/04/13  14:56:29  Mike_Rickabaugh]
 * 
 * Revision 1.1.2.2  92/02/20  17:04:30  Mike_Rickabaugh
 * 	Initial 3.10 version.
 * 	[92/02/20  16:50:35  Mike_Rickabaugh]
 * 
 * $EndLog$
 */

/*
 * Signed division
 *
 * On input:
 *	$23/t9  = return address (RA)
 *	$24/t10 = dividend (numerator) (A)
 *	$25/t11 = divisor (denominator) (B)
 *
 * On Return:
 *	$27/t12 = remainder (Q)
 *	All other registers, except $at and $gp, must be preserved
 */

#define ENTRY \
	.globl	__remq; \
	.ent	__remq; \
__remq: ; \
	ldgp	gp, 0(pv); \
	lda	sp, -8*8(sp); \
	stq	RA, 0*8(sp); \
	stq	t0, 1*8(sp); \
	stq	t1, 2*8(sp); \
	stq	t2, 3*8(sp); \
	stq	t3, 4*8(sp); \
	stq	a0, 5*8(sp); \
	stq	A, 6*8(sp); \
	stq	B, 7*8(sp); \
	.prologue 1; \
	/* check for LONG_MIN / -1 */; \
	addq	B, 1, t0	/* 0 if B == -1; != 0 otherwise */; \
	bis	A, A, t1	/* copy dividend */; \
	cmovne	t0, 0, t1	/* replace w/ 0 if divisor != -1 */; \
	subqv	zero, t1, t1	/* trap if dividend = LONG_MIN */; \
	subq	zero, A, t0	/* negate dividend */; \
	cmovlt	A, t0, A	/* get absolute value of dividend */; \
	subq	zero, B, t1	/* negate divisor */; \
	cmovlt	B, t1, B	/* get absolute value of divisor */

#define REMAINDER 1

#define RETURN \
	subq	zero, Q, AT; \
	ldq	A, 6*8(sp); \
	cmovlt	A, AT, Q; \
	ldq	RA, 0*8(sp); \
	ldq	t0, 1*8(sp); \
	ldq	t1, 2*8(sp); \
	ldq	t2, 3*8(sp); \
	ldq	t3, 4*8(sp); \
	ldq	a0, 5*8(sp); \
	ldq	B, 7*8(sp); \
	lda	sp, 8*8(sp); \
	ret	zero, (RA), 1

#define END	\
	.end	__remq

#include "divrem.S"
