/* syscall.S - generalized linux system call interface */

#include <linux/alpha/regdef.h>
#include <linux/alpha/pal.h>

/* Usage:
 *
 * int	syscall(syscall_number, arg1, arg2, arg3, arg4, arg5)
 *
 * syscall_number = the index of the system call we're invoking
 * arg1-arg5 = up to 5 integer arguments to the system call
 *
 * We need to do some arg shfting: the kernel expects the
 * syscall number in v0 and the first five args in a0-a4.
 *
 */

	.globl	syscall
	.ent	syscall
syscall:
	bis	a0,a0,v0	# Syscall number -> v0
	bis	a1,a1,a0	# arg1-arg5 -> a0-a4
	bis	a2,a2,a1
	bis	a3,a3,a2
	bis	a4,a4,a3
	bis	a5,a5,a4

	call_pal PAL_callsys	# Invoke system call
	beq	a3,1f		# br if a3=0 (no error)
	stl	v0,errno	# else store result in errno
	subq	zero,v0,v0	# and return -errno to caller
1:	ret	zero,(ra)
	.end	syscall


/* Functions called from the _syscallX macros.
 * Each call takes "x" number of args, followed by the syscall
 * number.  These routines are trivial dispatches.
 */

	.globl	syscall0
	.ent	syscall0
syscall0:
	bis		a0,a0,v0
	call_pal	PAL_callsys
	beq		a3,1f		# On error, return -errno to caller
	subq		zero,v0,v0
1:
	ret		zero,(ra)
	.end	syscall0

	.globl	syscall1
	.ent	syscall1
syscall1:
	bis		a1,a1,v0
	call_pal	PAL_callsys
	beq		a3,1f		# On error, return -errno to caller
	subq		zero,v0,v0
1:
	ret		zero,(ra)
	.end	syscall1

	.globl	syscall2
	.ent	syscall2
syscall2:
	bis		a2,a2,v0
	call_pal	PAL_callsys
	beq		a3,1f		# On error, return -errno to caller
	subq		zero,v0,v0
1:
	ret		zero,(ra)
	.end	syscall2

	.globl	syscall3
	.ent	syscall3
syscall3:
	bis		a3,a3,v0
	call_pal	PAL_callsys
	beq		a3,1f		# On error, return -errno to caller
	subq		zero,v0,v0
1:
	ret		zero,(ra)
	.end	syscall3

	.globl	syscall4
	.ent	syscall4
syscall4:
	bis		a4,a4,v0
	call_pal	PAL_callsys
	beq		a3,1f		# On error, return -errno to caller
	subq		zero,v0,v0
1:
	ret		zero,(ra)
	.end	syscall4

	.globl	syscall5
	.ent	syscall5
syscall5:
	bis		a5,a5,v0
	call_pal	PAL_callsys
	beq		a3,1f		# On error, return -errno to caller
	subq		zero,v0,v0
1:
	ret		zero,(ra)
	.end	syscall5

